#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# single point construction lines at an arbitrary angle
#

import math

import conobject
import tolerance
import util
import point
import maptree

class ACLine(conobject.Spcline):
    """A class for single point construction lines at a specified angle.

A ACLine object is derived from an Spcline,so it has
all that objects properties.

There is one additional attribute for an ACLine:

angle: A float value listing the angle at which this line rises or declines

The limits of the float value: -90.0 < value < 90.0. Any values outside
that range are adjusted to fall between those limits.

A ACLine has the following addtional methods:

{get/set}Angle(): Get/Set the angle of the ACLine.
mapPoint(): Test if a point is within some distance to an ACLine.
mapCoords(): Test if a coordinate pair is within some distance to an ACLine.
inRegion(): Return whether or not a ACLine passes through a bounded region.
clone(): Return an identical copy of an ACLine.
    """

    # class methods
    
    def __init__(self, p, a):
        """Initialize an ACLine object.

ACLine(p, a)

p: A Point object the line passes through
a: The angle at which the line rises or declines
        """
        _a = util.make_angle(a)
        conobject.Spcline.__init__(self, p)
        self.__angle = _a
        
    def __cmp__(self, obj):
        """Compare one ACLine object to another or to a Point.

The comparison of two ACLines is done by comparing the
slopes and y-intersections of the two lines. An ACLine
with infinite slope (angle is either 90.0 or -90.0) is
considered greater than any line with a slope.

For each line, calculate its slope(m) and y-intersection(b)

If A1 is vertical and A2 is not ==> return 1
If A1 is not vertical and A2 is ==> return -1

If both lines are vertical:
   Compare the "x" coordinates of each line's Point

   "x" of A1.point < "x" of A2.point ==> return -1
   "x" or A1.point > "x" of A2.point ==> return 1

   If we get here, return 0

Else:   
   b of A1 < b of A2 ==> return -1
   b of A1 > b of A2 ==> return 1

   If the y-intersections are identical, compare use the slope
   for comparison

   m of A1 < m of A2 ==> return -1
   m of A1 > m of A2 ==> return 1

   If we get here, return 0

Point comparisons always return 0.   
        """
        if not isinstance(obj, (point.Point, ACLine)):
            raise TypeError, "Invalid object for ACLine comparison: " + `obj`
        if isinstance(obj, point.Point):
            return 0
        _xs, _ys = self.getLocation().getCoords()
        _as = self.__angle
        _xo, _yo = obj.getLocation().getCoords()
        _ao = obj.getAngle()
        _sflag = False
        if abs(abs(_as) - 90.0) < 1e-6:
            _sflag = True
        _oflag = False
        if abs(abs(_ao) - 90.0) < 1e-6:
            _oflag = True
        if _sflag or _oflag:
            if _sflag and _oflag:
                if abs(_xs - _xo) < 1e-10:
                    _value = 0
                elif _xs > _xo:
                    _value = 1
                else:
                    _value = -1
            elif _sflag:
                _value = 1
            else:
                _value = -1
        else:
            _dtr = math.pi/180.0
            _ms = math.tan(_as * _dtr)
            _bs = _ys - (_ms * _xs)
            _mo = math.tan(_ao * _dtr)
            _bo = _yo - (_mo * _xo)
            if abs(_bs - _bo) < 1e-10:
                if abs(_ms - _mo) < 1e-10:
                    _value = 0
                elif _ms > _mo:
                    _value = 1
                else:
                    _value = -1
            elif _bs > _bo:
                _value = 1
            else:
                _value = -1
        return _value

    def __str__(self):
        _point = self.getLocation()
        _angle = self.__angle
        return "Angled construction line through %s at %g degrees" % (_point, _angle)

    def getAngle(self):
        """Return the angle of the ACLine.

getAngle()        
        """
        return self.__angle

    def setAngle(self, angle):
        """

setAngle(angle)

The argument a should be a float representing the angle
of the ACLine in degrees.
        """
        _a = util.make_angle(angle)
        self.modified()
        self.__angle = _a
    
    angle = property(getAngle, setAngle, None, "Angle of inclination.")

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the ACLine to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the ACLine. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _xs, _ys = self.getLocation().getCoords()
        _x2 = _xs + math.cos(self.__angle*(math.pi/180.0))
        _y2 = _ys + math.sin(self.__angle*(math.pi/180.0))
        _sqlen = pow((_x2 - _xs),2) + pow((_y2 - _ys),2)
        _r = ((_x - _xs)*(_x2 - _xs) + (_y - _ys)*(_y2 - _ys))/_sqlen
        _px = _xs + _r*(_x2 - _xs)
        _py = _ys + _r*(_y2 - _ys)
        _ap = None
        _xsep = abs(_px - _x)
        _ysep = abs(_py - _y)
        if ((_xsep < _t or abs(_xsep - _t) < 1e-10) and
            (_ysep < _t or abs(_ysep - _t) < 1e-10)):
            _ap = point.Point(_px, _py)
        return _ap

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the ACLine to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by Point object to an
actual Point on the ACLine. If the distance between the actual
Point and the Point used as an argument is less than the Tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _x, _y = _p.getCoords()
        return self.mapCoords(_x, _y, _t)

    def getProjection(self, x, y):
        """Find the projection point of some coordinates on the ACLine.

getProjection(x, y)

Arguments "x" and "y" should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _x1, _y1 = self.getLocation().getCoords()
        _angle = self.__angle
        if abs(_angle) < 1e-10: # horizontal
            _px = _x
            _py = _y1
        elif abs(abs(_angle) - 90.0) < 1e-10: # vertical
            _px = _x1
            _py = _y
        else:
            _rangle = _angle * (math.pi/180.0)
            _dx = math.cos(_rangle)
            _dy = math.sin(_rangle)
            _sqlen = pow(_dx, 2) + pow(_dy, 2)
            _rn = ((_x - _x1) * _dx) + ((_y - _y1) * _dy)
            _r = _rn/_sqlen
            _px = _x1 + (_r * _dx)
            _py = _y1 + (_r * _dy)
        return _px, _py

    def inRegion(self, xmin, ymin, xmax, ymax):
        """Return whether or not an ACLine passes through a region.

inRegion(xmin, ymin, xmax, ymax)

The first four arguments define the boundary. The method
will return True if the ACLine passes through the boundary.
Otherwise the function will return False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x, _y = self.getLocation().getCoords()
        _angle = self.__angle
        _val = False
        if _xmin < _x < _xmax and _ymin < _y < _ymax:
            _val = True
        elif abs(abs(_angle) - 90.0) < 1e-10: # vertical line
            if _xmin < _x < _xmax:
                _val = True
        elif abs(_angle) < 1e-10: # horizontal line
            if _ymin < _y < _ymax:
                _val = True
        else:
            #
            # the ACLine can be parameterized as
            #
            # x = u * (x2 - x1) + x1
            # y = u * (y2 - y1) + y1
            #
            # for u = 0, x => x1, y => y1
            # for u = 1, x => x2, y => y2
            #
            # if the ACLine passes through the region then there
            # will be valid u occuring at the region boundary
            #
            _rangle = _angle * (math.pi/180.0)
            _dx = math.cos(_rangle)
            _dy = math.sin(_rangle)
            #
            # x = xmin
            #
            _u = (_xmin - _x)/_dx
            _yt = (_u * _dy) + _y
            if (_ymin - 1e-10) < _yt < (_ymax + 1e-10): # catch endpoints
                _val = True
            if not _val:
                #
                # x = xmax
                #
                _u = (_xmax - _x)/_dx
                _yt = (_u * _dy) + _y
                if (_ymin - 1e-10) < _yt < (_ymax + 1e-10): # catch endpoints
                    _val = True
                if not _val:
                    #
                    # y = ymin
                    #
                    # if this fails there is no way the ACLine can be in
                    # region because it cannot pass through only one side
                    #
                    _u = (_ymin - _y)/_dy
                    _xt = (_u * _dx) + _x
                    if _xmin < _xt < _xmax:
                        _val = True
        return _val

    def clone(self):
        """Return a copy of a ACLine.

clone()
        """
        _loc = self.getLocation().clone()
        return ACLine(_loc, self.__angle)

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x, _y = self.getLocation().getCoords()
        _angle = self.__angle
        _coords = None
        if abs(abs(_angle) - 90.0) < 1e-10: # vertical
            if _xmin < _x < _xmax:
                _coords = (_x, _ymin, _x, _ymax)
        elif abs(_angle) < 1e-10: # horiztonal
            if _ymin < _y < _ymax:
                _coords = (_xmin, _y, _xmax, _y)
        else:
            #
            # the ACLine can be parameterized as
            #
            # x = u * (x2 - x1) + x1
            # y = u * (y2 - y1) + y1
            #
            # for u = 0, x => x1, y => y1
            # for u = 1, x => x2, y => y2
            #
            # The following is the Liang-Barsky Algorithm
            # for segment clipping modified slightly for
            # construction lines
            #
            _rangle = _angle * (math.pi/180.0)
            _dx = math.cos(_rangle)
            _dy = math.sin(_rangle)
            _P = [-_dx, _dx, -_dy, _dy]
            _q = [(_x - _xmin), (_xmax - _x), (_y - _ymin), (_ymax - _y)]
            _u1 = None
            _u2 = None
            _valid = True
            for _i in range(4):
                _pi = _P[_i]
                _qi = _q[_i]
                if abs(_pi) < 1e-10: # this should be caught earlier ...
                    if _qi < 0.0:
                        _valid = False
                        break
                else:
                    _r = _qi/_pi
                    if _pi < 0.0:
                        if _u2 is not None and _r > _u2:
                            _valid = False
                            break
                        if _u1 is None or _r > _u1:
                            _u1 = _r
                    else:
                        if _u1 is not None and _r < _u1:
                            _valid = False
                            break
                        if _u2 is None or _r < _u2:
                            _u2 = _r
            if _valid:
                _coords = (((_u1 * _dx) + _x),
                           ((_u1 * _dy) + _y),
                           ((_u2 * _dx) + _x),
                           ((_u2 * _dy) + _y))
        return _coords

class ACLineTree(maptree.PointMapTree):
    """A Tree class specific for ACLine objects.

This class is based on the PointMaptTree class, and
holds only ACLine objects.
    """
    def __init__(self):
        """Initialize a ACLineTree.

ACLineTree()
        """
        maptree.PointMapTree.__init__(self, ACLine)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a ACLine object held in a ACLineTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: A ACLine object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     ACLine objects are close enough to be considered equal.

If a ACLine object is found in the ACLineTree within the
tolerance, that object is returned. Otherwise, the function
returns None.
        """
        if not isinstance(obj, ACLine):
            raise TypeError, "Invalid type for searching in ACLineTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "CLineTree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _acl = self[_mid]
            _res = cmp(obj, _acl)
            if _res == 0:
                _ro = _acl
                break
            elif _res == 1:
                _lo = _mid + 1
            else:
                _hi = _mid
        if _ro is None:
            _ox, _oy = obj.getLocation().getCoords()
            _oa = obj.getAngle()
            _oflag = False
            _dtr = math.pi/180.0        
            if abs(abs(_oa) - 90.0) < 1e-6:
                _oflag = True
                _xmin = _ox - _t - 1e-10            
            else:
                _om = math.tan(_oa * _dtr)
                _ob = _oy - (_om * _ox)
            for _i in xrange(_mid, -1, -1):
                _sobj = self[_i]
                _sx, _sy = _sobj.getLocation().getCoords()
                _sa = _sobj.getAngle()
                _sflag = False
                if abs(abs(_sa) - 90.0) < 1e-6:
                    _sflag = True
                else:
                    _sm = math.tan(_sa * _dtr)
                    _sb = _sy - (_sm * _sx)
                if _oflag is False and _sflag is True:
                    break
                if _oflag is True and _sflag is False:
                    continue
                if _oflag and _sflag:
                    if _sx < _xmin:
                        break
                    _xdiff = abs(_ox - _sx)
                    if _xdiff < _t or abs(_xdiff - _t) < 1e-10:
                        _ro = _sobj
                        break
                else:
                    _bdiff = abs(_ob - _sb)
                    if _bdiff < _t or abs(_bdiff - _t) < 1e-10:
                        if abs(_om - _sm) < 1e-10:
                            _ro = _sobj
                            break
        return _ro

    def scan(self, obj, tol=tolerance.TOL):
        """Find a ACLine object held in a ACLineTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: An ACLine object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     ACLine object are close enough to be considered equal.

If a ACLine object is found in the ACLineTree within the
tolerance, that object is returned. Otherwise, the function
returns None.
        """
        if not isinstance(obj, ACLine):
            raise TypeError, "Invalid type for searching in ACLineTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "ACLineTree in modified state."
        _ox, _oy = obj.getLocation().getCoords()
        _oa = obj.getAngle()
        _oflag = False
        _dtr = math.pi/180.0        
        if abs(abs(_oa) - 90.0) < 1e-6:
            _oflag = True
            _xmax = _ox + _t + 1e-10            
        else:
            _om = math.tan(_oa * _dtr)
            _ob = _oy - (_om * _ox)
        _ro = None
        for _sobj in self:
            _sx, _sy = _sobj.getLocation().getCoords()
            _sa = _sobj.getAngle()
            _sflag = False
            if abs(abs(_sa) - 90.0) < 1e-6:
                _sflag = True
            else:
                _sm = math.tan(_sa * _dtr)
                _sb = _sy - (_sm * _sx)
            if _oflag is False and _sflag is True:
                break
            if _oflag is True and _sflag is False:
                continue
            if _oflag and _sflag:
                if _sx > _xmax:
                    break
                _xdiff = abs(_ox - _sx)
                if _xdiff < _t or abs(_xdiff - _t) < 1e-10:
                    _ro = _sobj
                    break
            else:
                _bdiff = abs(_ob - _sb)
                if _bdiff < _t or abs(_bdiff - _t) < 1e-10:
                    if abs(_om - _sm) < 1e-10:
                        _ro = _sobj
                        break
        return _ro

    def atCoords(self, angle, x, y):
        """Find an ACLine at a specific angle passing through some point

atCoords(angle, x, y)

The arguments for this method are:

angle: A float giving the angle of the ACLine
x: A float giving the a "x" coordinate the ACLine passes through.
y: A float giving the a "y" coordinate the ACLine passes through.


This method returns an ACLine if there is one in the tree
at the specified angle and passing through the specified
coordinates. Otherwise this method returns None.
        """
        _angle = util.make_angle(angle)
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        if self.isModified():
            raise ValueError, "ACLineTree in modified state."
        _vflag = False
        _dtr = math.pi/180.0
        if abs(abs(_angle) - 90.0) < 1e-6:
            _vflag = True
        if not _vflag:
            _m = math.tan(_angle * _dtr)
            _b = _y - (_m * _x)
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi:
            _mid = (_hi + _lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _acl = self[_mid]
            _ax, _ay = _acl.getLocation().getCoords()
            _aa = _acl.getAngle()
            _sflag = False
            if abs(abs(_aa) - 90.0) < 1e-6:
                _sflag = True
            else:
                _sm = math.tan(_aa * _dtr)
                _sb = _ay - (_sm * _ax)
            if _sflag and _vflag:
                if abs(_ax - x) < 1e-10:
                    _ro = _acl
                    break
                elif _ax < _x:
                    _lo = _mid + 1
                else:
                    _hi = _mid
            else:
                if _sflag or _vflag:
                    _hi = _mid
                else:
                    if abs(_aa - _angle) < 1e-10:
                        if abs(_sb - _b) < 1e-10:
                            _ro = _acl
                            break
                        elif _sb < _b:
                            _lo = _mid + 1
                        else:
                            _hi = _mid
                    elif _aa < _angle:
                        _lo = _mid + 1
                    else:
                        _hi = _mid
        return _ro

def intersect_region(acl, xmin, ymin, xmax, ymax):
    if not isinstance(acl, ACLine):
        raise TypeError, "Invalid ACLine: " + str(acl)
    _xmin = xmin
    if not isinstance(_xmin, float):
        _xmin = float(xmin)
    _ymin = ymin
    if not isinstance(_ymin, float):
        _ymin = float(ymin)
    _xmax = xmax
    if not isinstance(_xmax, float):
        _xmax = float(xmax)
    if _xmax < _xmin:
        raise ValueError, "Illegal values: xmax < xmin"
    _ymax = ymax
    if not isinstance(_ymax, float):
        _ymax = float(ymax)
    if _ymax < _ymin:
        raise ValueError, "Illegal values: ymax < ymin"
    _x, _y = acl.getLocation().getCoords()
    _angle = acl.getAngle()
    _x1 = _y1 = _x2 = _y2 = None
    if abs(abs(_angle) - 90.0) < 1e-6: # vertical
        if _xmin < _x < _xmax:
            _x1 = _x
            _y1 = _ymin
            _x2 = _x
            _y2 = _ymax
    elif abs(_angle) < 1e-6: # horizontal
        if _ymin < _y < _ymax:
            _x1 = _xmin
            _y1 = _y
            _x2 = _xmax
            _y2 = _y
    else:
        _dtr = math.pi/180.0
        _slope = math.tan(_angle * _dtr)
        _yint = _y - (_x * _slope)
        _xt = _x + math.cos(_angle * _dtr)
        _yt = _y + math.sin(_angle * _dtr)
        #
        # find y for x = xmin
        #
        _yt = (_slope * _xmin) + _yint
        if _ymin < _yt < _ymax:
            # print "hit at y for x=xmin"
            _x1 = _xmin
            _y1 = _yt
        #
        # find y for x = xmax
        #
        _yt = (_slope * _xmax) + _yint
        if _ymin < _yt < _ymax:
            # print "hit at y for x=xmax"
            if _x1 is None:
                _x1 = _xmax
                _y1 = _yt
            else:
                _x2 = _xmax
                _y2 = _yt
        if _x2 is None:
            #
            # find x for y = ymin
            #
            _xt = (_ymin - _yint)/_slope
            if _xmin < _xt < _xmax:
                # print "hit at x for y=ymin"
                if _x1 is None:
                    _x1 = _xt
                    _y1 = _ymin
                else:
                    _x2 = _xt
                    _y2 = _ymin
        if _x2 is None:
            #
            # find x for y = ymax
            #
            _xt = (_ymax - _yint)/_slope
            if _xmin < _xt < _xmax:
                # print "hit at x for y=ymax"
                if _x1 is None:
                    _x1 = _xt
                    _y1 = _ymax
                else:
                    _x2 = _xt
                    _y2 = _ymax
    return _x1, _y1, _x2, _y2
