#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# construction circle class
#

import point
import conobject
import circle
import tree
import maptree
import tolerance

class CCircle(circle._Circle, conobject._conobject, tree.TreeItem):
    """The visible and storable representation of _CCircle objects.

The CCircle class is derived from _Circle, _conobject,  and TreeItem
classes, so it shares all the attributes and methods from those classes.
    """
    def __init__(self, center, radius):
        """Initialize a CCircle object.

CCircle(center, radius)

There are two required parameters for instantiating a CCircle:

center: A Point object for the CCircle center
radius: A positive float value
        """
        tree.TreeItem.__init__(self)        
        conobject._conobject.__init__(self)
        circle._Circle.__init__(self, center, radius)
        self.reset()

    def __str__(self):
        _pos = self.getCenter().getCoords()
        _rad = self.getRadius()
        return "Construction Circle at %s with radius %g" % (_pos, _rad)
   
    def isModified(self):
        """Indicate if the CCircle has been changed.

isModified()

This method extends the TreeItem::isModified() method.
        """
        if self.getCenter().isModified():
            self.modified()
        return tree.TreeItem.isModified(self)

    def setCenter(self, c):
        """Set the center Point of the CCircle.

setCenter(c)

This method extends the _Circle::setCenter() method and sets
the modified flag indicating the CCircle has been altered.
        """
        if not isinstance(c, point.Point):
            raise TypeError, "Invalid center point: " + `c`
        circle._Circle.setCenter(self, c)
        self.modified()

    center = property(circle._Circle.getCenter, setCenter, None,
                      "CCircle center point.")
    
    def setRadius(self, r):
        """Set the radius of the CCircle.

setRadius(r)

This method extends the _Circle::setRadius() method and sets
the modified flag indicating the CCircle has been altered.
        """
        circle._Circle.setRadius(self, r)
        self.modified()

    radius = property(circle._Circle.getRadius, setRadius, None,
                      "CCircle radius.")

    def move(self, dx, dy):
        """Move a CCircle.

move(dx, dy)

This method extends the _Circle::move() method and sets
the modified flag indicating the CCircle has been altered.
        """
        circle._Circle.move(self, dx, dy)
        self.modified()

    def clone(self):
        """Create an identical copy of a CCircle.

clone()

This method overrides the _Circle::clone() method.
        """
        ccp = self.getCenter.clone()
        return CCircle(ccp, self.getRadius())

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the CCircle to the Point p.

mapPoint(p [, tol])

This method extends _Circle::mapPoint().
        """
        _pt = circle._Circle.mapPoint(self, p, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the CCircle to a coordinate pair.

mapCoords(x, y [, tol])

This method extends _Circle::mapCoords().
        """
        _pt = circle._Circle.mapCoords(self, x, y, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

class CCircleTree(maptree.PointMapTree):
    """A Tree class specific for CCircle objects.

This class is based on the Tree class, and holds only CCircle objects.
    """
    def __init__(self):
        """Initialize a CCircleTree.

CCircleTree()
        """
        maptree.PointMapTree.__init__(self, CCircle)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a CCircle object held in a CCircleTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: A CCircle object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a CCircle object is found in the CCircleTree within the
tolerance, the index of that object in the CCircleTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if not isinstance(_o, CCircle):
            raise TypeError, "Invalid type for searching in CCircleTree: " + `obj`

        _cc = maptree.PointMapTree.binscan(self, _o, tol)
        return _cc

    def scan(self, obj, tol=tolerance.TOL):
        """Find a CCircle object held in a CCircleTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: An CCircle object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a CCircle object is found in the CCircleTree within the
tolerance, the index of that object in the CCircleTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if not isinstance(_o, CCircle):
            raise TypeError, "Invalid type for searching in CCircleTree: " + `obj`
        _cc = maptree.PointMapTree.scan(self, _o, tol)
        return _cc
