#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# construction line/circle base class
#
# These variables provide the defaults for
# the construction line style attributes
#

import style
import linetype
import color
import tolerance
import point
import tree

class _conobject(object):
    """A base class for construction lines and circles.

This class is meant to provide the most basic bits for
construction lines and circles. All construction lines
and circles will share a common Style object so they'll
all be drawn with the same linetype, have the same color,
and be the same thickness (though they should never be
plotted out).

There is a single classmethod for _conobject objects:

getStyle(): Return the style object
    """

    # static class variables

    __defstyle = style.Style(u'Conobject Default Style',
                             linetype.Linetype(u'Construction Line', [2,2]),
                             color.Color(255, 0, 0),
                             0.05)

    def __init__(self):
        pass
            
    def getStyle(cls):
        return cls.__defstyle

    getStyle = classmethod(getStyle)

class _Spcline(object):
    """A base class for construction lines defined by a single point.

This class provides the basics for construction lines that
are defined at a single point, such as a horizontal or vertical
construction line. The class is derived from the _conobject class,
so it has that classes properties.

A _Spcline object has a single attribute:

point = A point object that the construction line passes through.

A _Spcline object has the following methods:

{get/set}Location(): Get/Set the point that defines the _Spcline.
move(): Move a _Spcline.
clone(): Return an identical copy of a _Spcline.
    """
    
    def __init__(self, p):
        """Instantiate a _Spcline object.

_Spcline(p)

The single argument should be a _Point object,
        """
        _p = p
        if not isinstance(_p, point._Point):
            _p = point._Point(p)
        _conobject.__init__(self)
        self.__point = _p

    def getLocation(self):
        """Return the point that the _Spcline passes through.

getLocation()        
        """
        return self.__point

    def setLocation(self, p):
        """Set the point that the _Spcline passes through.

setLocation(p)
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid _Point for location: " + str(p)
        self.__point = p

    location = property(getLocation, setLocation, None,
                        "Single point construction line Point.")

    def move(self, dx, dy):
        """Move a _Spcline.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__point.move(_dx, _dy)

    def clone(self):
        """Return a copy of a _Spcline.

clone()

This method should be overriden in any subclasses.
        """
        pass

class Spcline(_Spcline, _conobject, tree.TreeItem):
    """The visual and storable representation of _Spcline objects.

This class is the union of the _Spcline, _conobject, and
TreeItem classes, so it shares all the attributes and methods
of those classes.
    """
    def __init__(self, p):
        """Initialize a Spcline object.

Spcline(p)

The single argument should be a Point object,
        """
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _Spcline.__init__(self, _p)            
        _conobject.__init__(self)
        tree.TreeItem.__init__(self)
        _p.storeUser(self)

    def setLocation(self, p):
        """Set the point that the Spcline passes through.

setLocation(p)

This method extends the _Spcline::setLocation() method.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid _Point for location: " + str(p)
        _lp = self.getLocation()
        _Spcline.setLocation(self, p)
        self.modified()
        _lp.freeUser(self)
        p.storeUser(self)

    location = property(_Spcline.getLocation, setLocation, None,
                        "Single point construction line Point.")

    def move(self, dx, dy):
        """Move a Spcline.

move(dx, dy)

This method extends the _Spcline::move() method.
        """
        _Spcline.move(self, dx, dy)
        self.modified()
