#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for leader lines
#

import math
import array

import graphicobject
import tolerance
import style
import linetype
import color
import point
import baseobject
import util

class _Leader(object):
    """A class representing a leader line.

A leader line is usually used as a visual connector between
some text and an entity in a drawing. Leader lines cannot be
used to define an edge of some shape.

A _Leader object has the following attributes:
p1: A _Point object representing the first end point.
p2: A _Point object representing the leader mid point.
p3: A _Point object representing the final end point.
arrowsize: The size of the arrow at the end of the leader line

A _Leader object has the following methods:

getLeaderPoints(): Return the points defining the _Leader
{get/set}P1(): Get/Set the _Leader first endpoint.
{get/set}P2(): Get/Set the _Leader midpoint
{get/set}P3(): Get/Set the _Leader final endpoint
{get/set}ArrowSize(): Get/Set the _Leader line arrow size.
calcArrowPoints(): Calculate where the _Leader arrow points are.
getArrowPoints(): Return where the _Leader arrow points are.
move(): Move the _Leader.
inRegion(): Test if the _Leader is visible in some area.
clone(): Make an identical copy of a _Leader.
onLeader(): See if a point lies within some distance of a _Leader.
    """

    # member functions

    def __init__(self, p1, p2, p3, size=1.0):
        """Initialize a _Leader object.

_Leader(p1, p2, p3[, size])

The following arguments are required:

p1: _Leader first endpoint - may be a _Point or a two-item tuple of floats
p2: _Leader mid point - may be a _Point or a two-item tuple of floats
p3: _Leader final endpoint - may be a _Point or a two-item tuple of floats

Argument "size" is optional. It gives the size of the arrow at
the end of the leader line, and defaults to 1.0.
        """
        _p1 = p1
        if not isinstance(_p1, point._Point):
            _p1 = point._Point(p1)
        _p2 = p2
        if not isinstance(_p2, point._Point):
            _p2 = point._Point(p2)
        if _p1 == _p2:
            raise ValueError, "_Leader objects cannot have equivalent points."
        _p3 = p3
        if not isinstance(_p3, point.Point):
            _p3 = point._Point(_p3)
        if _p1 == _p3:
            raise ValueError, "_Leader objects cannot have equivalent points."
        if _p2 == _p3:
            raise ValueError, "_Leader objects cannot have equivalent points."
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid arrow size: %g" % _size
        self.__p1 = _p1
        self.__p2 = _p2
        self.__p3 = _p3
        self.__size = _size
        self.__arrow_pts = array.array('d', [0.0, 0.0, 0.0, 0.0])
        self.calcArrowPoints()

    def __str__(self):
        return "Leader: %s to %s to %s" % (self.__p1, self.__p2, self.__p3)

    def __eq__(self, obj):
        """Compare two leader lines for equality.
        """
        if not isinstance(obj, _Leader):
            raise TypeError, "Invalid leader comparison: " + `obj`
        _sp1 = self.__p1
        _sp2 = self.__p2
        _sp3 = self.__p3
        _p1, _p2, _p3 = obj.getLeaderPoints()
        _res = False
        if (_sp1 == _p1 and _sp2 == _p2 and _sp3 == _p3):
            _res = True
        return _res

    def __ne__(self, obj):
        """Compare two leader lines for inequality.
        """
        if not isinstance(obj, _Leader):
            raise TypeError, "Invalid leader comparison: " + `obj`
        _sp1 = self.__p1
        _sp2 = self.__p2
        _sp3 = self.__p3
        _p1, _p2, _p3 = obj.getLeaderPoints()
        _res = True
        if (_sp1 == _p1 and _sp2 == _p2 and _sp3 == _p3):
            _res = False
        return _res

    def getLeaderPoints(self):
        """Get the points defining the _Leader.

getLeaderPoints()

This function returns a tuple containing the three _Point objects
that are the defining points of the _Leader
        """
        return (self.__p1, self.__p2, self.__p3)

    def getP1(self):
        """Return the first endpoint _Point of the _Leader

getP1()
        """
        return self.__p1

    def setP1(self, p):
        """Set the first endpoint _Point of the _Leader

setP1(p)

The argument "p" should be a _Point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid _Point for p1 endpoint: " + `p`
        self.__p1 = p

    p1 = property(getP1, setP1, None, "First endpoint of the Leader.")

    def getP2(self):
        """Return the midpoint _Point of the _Leader.

getP2()
        """
        return self.__p2

    def setP2(self, p):
        """Set the midpoint _Point of the _Leader.

setP2(p)

The argument "p" should be a _Point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid Point for p2 endpoint: " + `p`
        self.__p2 = p

    p2 = property(getP2, setP2, None, "Leader midpoint.")

    def getP3(self):
        """Return the final _Point of the _Leader.

getP3()
        """
        return self.__p3

    def setP3(self, p):
        """Set the final endpoint _Point of the _Leader.

setP3(p)

The argument "p" should be a _Point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid Point for p3 endpoint: " + `p`
        self.__p3 = p

    p3 = property(getP3, setP3, None, "Third endpoint of the Leader.")

    def getArrowSize(self):
        """Return the size of the leader line arrow.

getArrowSize()
        """
        return self.__size

    def setArrowSize(self, size):
        """Set the size of the leader line arrow.

setSize(size)

Argument "size" should be a float greater than or equal to 0.0.
        """
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid arrow size: %g" % _size
        self.__size = _size
        self.calcArrowPoints()

    arrowsize = property(getArrowSize, setArrowSize,
                    None, "_Leader line arrow size.")

    def calcArrowPoints(self):
        """Calculate where the _Leader arrow points are.

calcArrowPoints()
        """
        _x1, _y1 = self.__p2.getCoords()
        _x2, _y2 = self.__p3.getCoords()
        if abs(_x2 - _x1) < 1e-10: # vertical
            _cosine = 0.0
            if _y2 > _y1:
                _sine = 1.0
            else:
                _sine = -1.0

        elif abs(_y2 - _y1) < 1e-10: # horizontal
            _sine = 0.0
            if _x2 > _x1:
                _cosine = 1.0
            else:
                _cosine = -1.0
        else:
            _angle = math.atan2((_y2 - _y1), (_x2 - _x1))
            _sine = math.sin(_angle)
            _cosine = math.cos(_angle)
        _size = self.__size
        _height = _size/5.0
        # p1 -> (x,y) = (-size, _height)
        self.__arrow_pts[0] = (_cosine * (-_size) - _sine * _height) + _x2
        self.__arrow_pts[1] = (_sine * (-_size) + _cosine * _height) + _y2
        # p2 -> (x,y) = (-size, -_height)
        self.__arrow_pts[2] = (_cosine * (-_size) - _sine *(-_height)) + _x2
        self.__arrow_pts[3] = (_sine * (-_size) + _cosine *(-_height)) + _y2

    def getArrowPoints(self):
        """Return the endpoints of the _Leader arrow.

getArrowPoints()

This method returns an array holding four float values.
        """
        return self.__arrow_pts

    def move(self, dx, dy):
        """Move a _Leader.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__p1.move(_dx, _dy)
        self.__p2.move(_dx, _dy)
        self.__p3.move(_dx, _dy)
        self.calcArrowPoints()

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest _Point on the _Leader to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual _Point on the _Leader. If the distance between the actual
_Point and the coordinates used as an argument is less than the tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _x3, _y3 = self.__p3.getCoords()
        _ap = None
        _pt = util.map_coords(_x, _y, _x1, _y1, _x2, _y2, _t)
        if _pt is None:
            _pt = util.map_coords(_x, _y, _x2, _y2, _x3, _y3, _t)
        if _pt is not None:
            _px, _py = _pt
            _ap = point._Point(_px, _py)
        return _ap

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest _Point on the _Leader to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A _Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by _Point object to an
actual _Point on the _Leader. If the distance between the actual
_Point and the _Point used as an argument is less than the tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point._Point):
            _p = point._Point(p)
        _t = tolerance.toltest(tol)
        _x, _y = _p.getCoords()
        return self.mapCoords(_x, _y, _t)

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a _Leader exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the _Leader lies within that area. If
the optional argument "fully" is used and is True, then all
the _Leader points must lie within the boundary. Otherwise,
the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        # print "xmin: %g" % _xmin
        # print "ymin: %g" % _ymin
        # print "xmax: %g" % _xmax
        # print "ymax: %g" % _ymax
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _x3, _y3 = self.__p3.getCoords()
        # print "x1: %g" % _x1
        # print "y1: %g" % _y1
        # print "x2: %g" % _x2
        # print "y2: %g" % _y2
        # print "x3: %g" % _x3
        # print "y3: %g" % _y3
        _pxmin = min(_x1, _x2, _x3)
        _pymin = min(_y1, _y2, _y3)
        _pxmax = max(_x1, _x2, _x3)
        _pymax = max(_y1, _y2, _y3)
        # print "pxmin: %g" % _pxmin
        # print "pymin: %g" % _pymin
        # print "pxmax: %g" % _pxmax
        # print "pymax: %g" % _pymax
        if (_xmin < _pxmin and _xmax > _pxmax and
            _ymin < _pymin and _ymax > _pymax):
            return True
        _val = False
        if not fully:
            _val = util.in_region(_x1, _y1, _x2, _y2,
                                  _xmin, _ymin, _xmax, _ymax)
            if not _val:
                _val = util.in_region(_x2, _y2, _x3, _y3,
                                      _xmin, _ymin, _xmax, _ymax)
        return _val

    def clone(self):
        """Create an identical copy of a _Leader.

clone()
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        _cp3 = self.__p3.clone()
        _size = self.__size
        return _Leader(_cp1, _cp2, _cp3, _size)

class Leader(_Leader, baseobject.ModObject, graphicobject.GraphicObject):
    """The storable and displayed representation of _Leader objects.

The Leader class is derived from _Leader, treeItem, and
GraphicObject classes, so it shares all the attributes and
methods of those classes.
    """

    #
    # The default style for the Leader class
    #
    
    __defstyle = style.Style(u'Leader Default Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    def __init__(self, p1, p2, p3, size=1.0,
                 st=None, lt=None, col=None, t=None):
        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 == _p2:
            raise ValueError, "Leader objects cannot have equivalent points."
        _p3 = p3
        if not isinstance(_p3, point.Point):
            _p3 = point.Point(p3)
        if _p1 == _p3:
            raise ValueError, "Leader objects cannot have equivalent points."
        if _p2 == _p3:
            raise ValueError, "Leader objects cannot have equivalent points."
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid Leader arrow size: %g" % _size
        _st = st
        if _st is None:
            _st = Leader.__defstyle
        baseobject.ModObject.__init__(self)
        _Leader.__init__(self, _p1, _p2, _p3, _size)
        graphicobject.GraphicObject.__init__(self, _st, lt, col, t)
        _p1.storeUser(self)
        _p2.storeUser(self)
        _p3.storeUser(self)

    def isModified(self):
        """Indicate if the Leader has been changed.

isModified()

This method extends the ModObject::isModified() method.
        """
        _p1, _p2, _p3 = self.getLeaderPoints()
        if _p1.isModified() or _p2.isModified() or _p3.isModified():
            self.modified()
        return baseobject.ModObject.isModified(self)

    def setP1(self, p):
        """Set the first endpoint Point of the Leader

setP1(p)

This method extends _Leader::setP1() and sets the modified
flag indicating that the Leader has been altered.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _p1 = self.getP1()
        _Leader.setP1(self, p)
        self.modified()
        _p1.freeUser(self)
        p.storeUser(self)

    p1 = property(_Leader.getP1, setP1, None,
                  "First endpoint of the Leader.")

    def setP2(self, p):
        """Set the midpoint Point of the Leader.

setP2(p)

This method extends _Leader::setP2() and sets the modified
flag indicating that the Leader has been altered.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _p2 = self.getP2()
        _Leader.setP2(self, p)
        self.modified()
        _p2.freeUser(self)
        p.storeUser(self)

    p2 = property(_Leader.getP2, setP2, None,
                  "Midpoint of the Leader.")

    def setP3(self, p):
        """Set the final endpoint Point of the Leader.

setP3(p)

This method extends _Leader::setP3() and sets the modified
flag indicating that the Leader has been altered.
        """
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _p3 = self.getP3()
        _Leader.setP3(self, p)
        self.modified()
        _p3.freeUser(self)
        p.storeUser(self)

    p3 = property(_Leader.getP3, setP3, None,
                  "Final endpoint of the Leader.")

    def move(self, dx, dy):
        """Move a Leader.

move(dx, dy)

This method extends _Leader::move() and sets the modified
flag indicating that the Leader has been altered.
        """
        _Leader.move(self, dx, dy)
        self.modified()

    def clone(self):
        """Create an identical copy of a Leader.

clone()

This method overrides the _Leader::clone() method.
        """
        _p1, _p2, _p3 = self.getLeaderPoints()
        _cp1 = _p1.clone()
        _cp2 = _p2.clone()
        _cp3 = _p3.clone()
        _size = self.getArrowSize()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Leader(_cp1, _cp2, _cp3, _size, _st, _lt, _col, _th)

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Leader to the point p.

mapPoint(p [, tol])

This function extends _Leader::mapPoint().
        """
        _pt = _Leader.mapPoint(self, p, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Leader to a coordinate pair.

mapCoords(x, y[, tol])

This function extends _Leader::mapCoords().
        """
        _pt = _Leader.mapCoords(self, x, y, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt
